# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Saturation Environment
=======================
"""
import dataclasses
import os
import random
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import orjson
from gym import Env, spaces
from tptp_lark_parser.grammar import Clause
from tptp_lark_parser.tptp_parser import TPTPParser

from gym_saturation.clause_space import ClauseSpace
from gym_saturation.logic_ops.factoring import all_possible_factors
from gym_saturation.logic_ops.paramodulation import all_paramodulants_from_list
from gym_saturation.logic_ops.reflexivity_resolution import (
    all_possible_reflexivity_resolvents,
)
from gym_saturation.logic_ops.resolution import all_possible_resolvents
from gym_saturation.logic_ops.utils import (
    is_tautology,
    reduce_to_proof,
    reindex_variables,
)

STATE_DIFF_UPDATED = "state_diff_updated"
POSITIVE_ACTIONS = "positive_actions"
PROBLEM_FILENAME = "problem_filename"
MAX_CLAUSES = 100000


class SaturationEnv(Env):
    """
    Saturation algorithm defined in a Reiforcement Learning friendly way.

    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> tptp_folder = files("gym_saturation").joinpath(
    ...     os.path.join("resources", "TPTP-mock")
    ... )
    >>> from glob import glob
    >>> problem_list = sorted(glob(
    ...     os.path.join(tptp_folder, "Problems", "*", "*1-1.p")
    ... ))
    >>> env = SaturationEnv(problem_list)
    >>> env.seed(0)
    0
    >>> len(env.reset()["real_obs"])
    4

    one can look at the current state in TPTP format

    >>> print(env.render())
    cnf(this_is_a_test_case_1, hypothesis, p(c)).
    cnf(this_is_a_test_case_2, hypothesis, ~p(c)).
    cnf(test_axiom, axiom, c = d).
    cnf(test_axiom_2, axiom, ~c = e).

    ``ansi`` mode returns a JSON representation of the state
    it should be more easily parsable than TPTP, although less human-friendly

    >>> env.render("ansi")
    [...{"literals":[{"negated":false,"atom":{"index":298,"arguments":[{"ind...

    other modes are not implemented yet

    >>> env.render(mode="rgb_array")
    Traceback (most recent call last):
     ...
    NotImplementedError

    the test theorem can be proved in three steps

    >>> next_state, reward, done, info = env.step(0)

    ``info`` dict contains the state diff, for example

    >>> import json
    >>> json.loads(info["state_diff_updated"][0])["processed"]
    True

    repeating actions is not allowed

    >>> env.step(0)
    Traceback (most recent call last):
     ...
    ValueError: action 0 is not valid

    there is no reward until the end of an episode

    >>> (reward, done)
    (0.0, False)

    if a proof is found, then reward is ``+1``

    >>> env.step(1)[1:3]
    (1.0, True)

    TSTP proof is now available (one can add ``include`` directive before it
    for validation purposes)

    >>> print(env._tptp_parser.cnf_parser.pretty_print(
    ...     env._tptp_parser.parse(env.tstp_proof)[0])
    ... )
    cnf(..., lemma, $false, inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).

    the relevant actions are filtered too

    >>> env.positive_actions
    (0, 1, 4)

    the total number of clauses in the state is limited by the ``max_clauses``
    parameter. Let's try setting it and repeating the same solution of the same
    problem:

    >>> env = SaturationEnv(problem_list, max_clauses=3)
    >>> _ = env.seed(0)
    >>> old_obs = env.reset()

    after the first step we bypass ``max_clauses`` by one, so the episode
    finishes with failure:

    >>> obs, reward, done, _ = env.step(0)
    >>> done, reward
    (True, 0.0)
    """

    metadata = {"render_modes": ["ansi", "human"]}
    reward_range = (0, 1)

    action_space: spaces.Discrete
    observation_space: spaces.Dict

    def __init__(
        self,
        problem_list: List[str],
        max_clauses: int = MAX_CLAUSES,
    ):
        """
        Initialize spaces et al.

        :param problem_list: a list of the names of TPTP problem files
        :param max_clauses: maximal number of clauses to store in proof state
        """
        super().__init__()
        self.problem_list = problem_list
        self._state: Dict[str, Clause] = {}
        self._state_set: Set[Tuple[bytes, ...]] = set()
        self.action_space = spaces.Discrete(max_clauses)
        self.observation_space = spaces.Dict(
            {
                "action_mask": spaces.Box(0, 1, shape=(max_clauses,)),
                "real_obs": ClauseSpace(),
            }
        )
        self.problem: Optional[str] = None
        tptp_folder = os.path.join(
            os.path.dirname(problem_list[0]), "..", ".."
        )
        self._tptp_parser = TPTPParser(tptp_folder, extendable=True)

    def _init_clauses(self) -> Dict[str, Clause]:
        self.problem = random.choice(self.problem_list)
        with open(self.problem, "r", encoding="utf-8") as problem_file:
            problem_text = problem_file.read()
        parsed_clauses = self._tptp_parser.parse(problem_text)
        return {
            clause.label: dataclasses.replace(
                clause,
                birth_step=0,
                inference_parents=(),
                inference_rule=None,
                processed=False,
            )
            for clause in parsed_clauses
        }

    def reset(self) -> Union[dict, Tuple[dict, dict]]:  # noqa: D102
        self._state = reindex_variables(self._init_clauses())
        self._state_set = set(
            map(
                lambda clause: tuple(
                    sorted(map(orjson.dumps, clause.literals))
                ),
                self._state.values(),
            )
        )
        return self.state

    def _add_to_state(self, new_clauses: Tuple[Clause, ...]) -> None:
        birth_step = 1 + self.last_birth_step
        for clause in new_clauses:
            if not is_tautology(clause):
                sorted_literals = tuple(
                    sorted(map(orjson.dumps, clause.literals))
                )
                if sorted_literals not in self._state_set:
                    self._state[clause.label] = dataclasses.replace(
                        clause, birth_step=birth_step, processed=False
                    )
                    self._state_set.add(sorted_literals)

    def _do_deductions(self, action: int) -> Tuple[bytes, ...]:
        state_len_before = len(self._state)
        given_clause_label, given_clause = list(self._state.items())[action]
        unprocessed_clauses = tuple(
            clause for clause in self._state.values() if clause.processed
        )
        self._add_to_state(
            all_possible_resolvents(
                unprocessed_clauses,
                given_clause,
            )
        )
        self._add_to_state(
            all_paramodulants_from_list(
                unprocessed_clauses,
                given_clause,
            )
        )
        self._add_to_state(
            all_possible_factors(
                given_clause,
            )
        )
        self._add_to_state(
            all_possible_reflexivity_resolvents(
                given_clause,
            )
        )
        self._state[given_clause_label] = dataclasses.replace(
            given_clause, processed=True
        )
        return tuple(
            map(orjson.dumps, list(self._state.values())[state_len_before:])
        ) + (orjson.dumps(list(self._state.values())[action]),)

    def _proof_found_result(
        self, reward: float, info: Dict[str, Any]
    ) -> Tuple[float, bool, Dict[str, Any]]:
        if tuple(
            True for clause in self._state.values() if clause.literals == ()
        ):
            info[POSITIVE_ACTIONS] = self.positive_actions
            return 1.0, True, info
        return reward, False, info

    def _max_clauses_result(
        self,
        done: bool,
        info: Dict[str, Any],
    ) -> Tuple[bool, Dict[str, Any]]:
        if not done:
            if len(self._state) > self.action_space.n:
                info.pop(STATE_DIFF_UPDATED)
                return True, info
        return done, info

    def step(self, action: int) -> Tuple[dict, float, bool, Dict[str, Any]]:
        # noqa: D301
        """
        Run one timestep of the environment's dynamics.

        When end of episode is reached, you are responsible for calling
        ``reset()`` to reset this environment's state.
        Accepts an action and returns a tuple (observation, reward, done, info)

        :param action: an action provided by the agent
        :returns: a tuple of four values:\n
            * observation: agent's observation of the current environment
            * reward: amount of reward returned after previous action
            * done: whether the episode has ended, in which case further
              ``step()`` calls will return undefined results
            * info: contains auxiliary diagnostic information (helpful for
              debugging, and sometimes learning)
        :raises ValueError: if the ``action`` identifies an already processed
            clause
        """
        if list(self._state.values())[action].processed:
            raise ValueError(f"action {action} is not valid")
        updated = self._do_deductions(action)
        reward = 0.0
        info = {STATE_DIFF_UPDATED: updated, PROBLEM_FILENAME: self.problem}
        reward, done, info = self._proof_found_result(reward, info)
        done |= min(
            False if clause.processed is None else clause.processed
            for clause in self._state.values()
        )
        done, info = self._max_clauses_result(done, info)
        return self.state, reward, done, info

    @property
    def last_birth_step(self) -> int:
        """Return the last birth step number of clauses in the proof state."""
        return max(getattr(clause, "birth_step", 0) for clause in self._state)

    # pylint: disable=inconsistent-return-statements
    def render(self, mode="human"):  # noqa: D102
        if mode == "ansi":
            return self.state["real_obs"]
        if mode == "human":
            return "\n".join(
                map(
                    self._tptp_parser.cnf_parser.pretty_print,
                    self._state.values(),
                )
            )
        super().render(mode=mode)

    @property
    def state(self) -> dict:
        """Return environment state in Python ``dict`` format."""
        return {
            "real_obs": [
                orjson.dumps(clause) for clause in self._state.values()
            ],
            "action_mask": (
                np.array(
                    [
                        0.0 if clause.processed else 1.0
                        for clause in self._state.values()
                    ]
                    + self.action_space.n * [0.0],
                    np.float32,
                )
            )[: self.action_space.n],
        }

    def seed(self, seed=None):  # noqa: D102
        random.seed(seed)
        return seed

    @property
    def tstp_proof(self) -> str:
        """Return TSTP proof (if found; raises an error otherwise)."""
        return "\n".join(
            reversed(
                [
                    self._tptp_parser.cnf_parser.pretty_print(clause)
                    for clause in reduce_to_proof(self._state)
                    if clause.inference_rule is not None
                ]
            )
        )

    @property
    def positive_actions(self) -> Tuple[int, ...]:
        """
        Return a sequence of actions which contributed to the proof found.

        If there is no proof yet, raises an error.
        """
        proof = reduce_to_proof(self._state)
        return tuple(
            action
            for action, clause in enumerate(self._state.values())
            if clause in proof
        )



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Logic Operations Utils
=======================
"""
import dataclasses
from itertools import chain
from operator import attrgetter
from typing import Any, Dict, Tuple, Union

from tptp_lark_parser.grammar import (
    EQUALITY_SYMBOL_ID,
    Clause,
    Function,
    Predicate,
    Proposition,
    Term,
    Variable,
)

from gym_saturation.logic_ops.substitution import Substitution


class WrongRefutationProofError(Exception):
    """Exception raised when proof is requested but not found yet."""


def is_subproposition(one: Proposition, two: Proposition) -> bool:
    """
    Check whether proposition ``one`` is part of a proposition ``two``.

    :param one: presumably shorter proposition which is probably a part
    :param two: presumably longer proposition which is probably the whole
    :returns: whether ``one`` is a part of ``two``
    """
    if isinstance(two, Variable):
        return one == two
    if isinstance(two, (Function, Predicate)):
        for argument in two.arguments:
            if is_subproposition(one, argument):
                return True
    return False


def get_variable_list(
    clause: Union[Clause, Proposition]
) -> Tuple[Variable, ...]:
    """
    Find all variables present in a clause.

    >>> from tptp_lark_parser.grammar import Literal
    >>> get_variable_list(Clause((Literal(False,
    ...     Predicate(0, (Function(0, (Variable(0), Variable(0))),))
    ... ),)))
    (Variable(index=0), Variable(index=0))

    :param clause: a clause
    :returns: a list (with repetitions) of variables from there clause
    """
    variable_list: Tuple[Variable, ...] = ()
    if isinstance(clause, Clause):
        for literal in clause.literals:
            for term in literal.atom.arguments:
                variable_list = variable_list + get_variable_list(term)
    elif isinstance(clause, Function):
        for term in clause.arguments:
            variable_list = variable_list + get_variable_list(term)
    elif isinstance(clause, Variable):
        variable_list = variable_list + (clause,)
    return variable_list


def _shift_variables(
    clauses: Dict[str, Clause], variable_list: Tuple[Variable, ...], shift: int
) -> Dict[str, Clause]:
    new_clauses: Dict[str, Clause] = {}
    for label, clause in clauses.items():
        new_clause = clause
        for i, variable in enumerate(variable_list):
            new_clause = Substitution(
                variable, Variable(shift + i)
            ).substitute_in_clause(new_clause)
        new_clauses[label] = new_clause
    return new_clauses


def reindex_variables(clauses: Dict[str, Clause]) -> Dict[str, Clause]:
    """
    Rename variables so that each clause has its unique set of variables.

    >>> from tptp_lark_parser.grammar import Literal
    >>> clause = Clause((Literal(False,
    ...     Predicate(0, (Variable(2), Variable(1), Variable(0)))
    ... ),))
    >>> sorted(map(
    ...     attrgetter("index"),
    ...     reindex_variables(
    ...         {"this_is_a_test_case": clause}
    ...     )["this_is_a_test_case"].literals[0].atom.arguments
    ... ))
    [0, 1, 2]
    >>> clause = Clause((Literal(False,
    ...     Predicate(0, (Variable(5), Variable(10), Variable(5)))
    ... ),))
    >>> sorted(map(
    ...     attrgetter("index"),
    ...     reindex_variables(
    ...         {"this_is_a_test_case": clause}
    ...     )["this_is_a_test_case"].literals[0].atom.arguments
    ... ))
    [0, 1, 1]

    :param clauses: a map of clause labels to clauses
    :returns: the list of clauses with renamed variables
    """
    variable_list = _flat_list(tuple(map(get_variable_list, clauses.values())))
    shift = max(
        len(variable_list),
        1 + max(map(attrgetter("index"), variable_list), default=-1),
    )
    new_clauses = _shift_variables(clauses, variable_list, shift)
    variable_list = _flat_list(
        tuple(map(get_variable_list, new_clauses.values()))
    )
    new_clauses = _shift_variables(new_clauses, variable_list, 0)
    return new_clauses


def is_tautology(clause: Clause) -> bool:
    """
    Check whether there are two literals (negated and not) with the same atom.

    >>> from tptp_lark_parser.grammar import Literal
    >>> is_tautology(Clause((Literal(False, Predicate(7, ())),)))
    False
    >>> is_tautology(Clause(
    ...     (Literal(False, Predicate(7, ())), Literal(True, Predicate(7, ())))
    ... ))
    True
    >>> is_tautology(Clause(
    ...     (Literal(False, Predicate(1, (Variable(0), Variable(0)))),)
    ... ))
    True

    :param clause: a clause to check
    :returns: whether the clause is a primitive tautology or not
    """
    for i, literal in enumerate(clause.literals):
        for j, another_literal in enumerate(clause.literals):
            if (
                i != j
                and literal.negated != another_literal.negated
                and literal.atom == another_literal.atom
            ):
                return True
        if literal.atom.index == EQUALITY_SYMBOL_ID and (
            literal.atom.arguments[0] == literal.atom.arguments[1]
        ):
            return True
    return False


def clause_length(clause: dict) -> int:
    """
    Find the length of arguments of each predicate.

    Negation adds one to each literal.

    :param clause: a clause in JSON representation
    :return: sctructural length of a clause

    >>> from tptp_lark_parser.grammar import Literal
    >>> import orjson
    >>> clause_length(orjson.loads(orjson.dumps(
    ...     Clause((Literal(True, Predicate("p", (Function("this_is_a_test_case", ()),))),))
    ... )))
    3
    """
    length = 0
    if isinstance(clause, dict):
        for key, value in clause.items():
            if key in {"negated", "index"}:
                length += 1
            if isinstance(value, dict):
                length += clause_length(value)
            if isinstance(value, (list, tuple)):
                for item in value:
                    length += clause_length(item)
    return length


def proposition_length(proposition: Proposition) -> int:
    """
    Find the number of functional, predicate and variable symbols.

    :param proposition: a function, a predicate or a variable
    :return: sctructural length of a proposition

    >>> proposition_length(Predicate(7, (Function(0, (Variable(0),)),)))
    3
    """
    length = 0
    if isinstance(proposition, Variable):
        return 1
    for subterm in proposition.arguments:
        length += proposition_length(subterm)
    return 1 + length


def clause_in_a_list(clause: Clause, clauses: Tuple[Clause, ...]) -> bool:
    """
    Check whether a clause is in a list.

    >>> clause_in_a_list(Clause((), label="one"), (Clause((), label="two"),))
    True

    :param clause: some clause
    :param clauses: a list of clauses
    :returns: whether in the list there is a clause with a literals set to a
        given clause
    """
    return clause.literals in set(map(lambda clause: clause.literals, clauses))


class NoSubtermFound(Exception):
    """Sometimes a subterm index is larger than term length."""


def subterm_by_index(atom: Proposition, index: int) -> Term:
    """
    Extract a subterm using depth-first search.

    >>> atom = Predicate(7, (
    ...     Function(0, (Variable(0),)), Function(1, (Variable(1),))
    ... ))
    >>> subterm_by_index(atom, 0)
    Traceback (most recent call last):
     ...
    ValueError: subterm with index 0 exists only for terms, but got: Predica...
    >>> subterm_by_index(atom, 1) == atom.arguments[0]
    True
    >>> subterm_by_index(atom, 2) == atom.arguments[0].arguments[0]
    True
    >>> subterm_by_index(atom, 4) == atom.arguments[1].arguments[0]
    True

    :param atom: a predicate or a term
    :param index: an index of a desired subterm
    :returns: a subterm
    :raises ValueError: when trying to get a term with index 0 of a predicate
    :raises NoSubtermFound: if subterm with a given index doesn't exist
    """
    if index == 0:
        if isinstance(atom, (Function, Variable)):
            return atom
        raise ValueError(
            f"subterm with index 0 exists only for terms, but got: {atom}"
        )
    subterm_length = 1
    if not isinstance(atom, Variable):
        for argument in atom.arguments:
            try:
                return subterm_by_index(argument, index - subterm_length)
            except NoSubtermFound as error:
                subterm_length += error.args[0]
    raise NoSubtermFound(subterm_length)


class CantReplaceTheWholeTerm(Exception):
    """An exception raised when trying to replace a subterm with index 0."""


class TermSelfReplace(Exception):
    """An exception raised when trying to replace a subterm with itself."""


def _replace_if_not_the_same(old_term: Term, new_term: Term) -> Term:
    if old_term == new_term:
        raise TermSelfReplace
    return new_term


def replace_subterm_by_index(
    atom: Proposition, index: int, term: Term
) -> Proposition:
    """
    Replace a subterm with a given index (depth-first search) by a new term.

    >>> atom = Predicate(7, (
    ...     Function(0, (Variable(0),)), Function(1, (Variable(2),))
    ... ))
    >>> replace_subterm_by_index(atom, 0, Variable(3))
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.utils.NoSubtermFound: 5
    >>> "this_is_a_test_case", replace_subterm_by_index(atom, 4, Function(2, (Variable(3),)))
    ('this_is_a_test_case', Predicate(index=7, arguments=(Function(index=0, arguments=(Variable(index=0),)), Function(index=1, arguments=(Function(index=2, arguments=(Variable(index=3),)),)))))
    >>> replace_subterm_by_index(Predicate(7, (Variable(0),)), 1, Variable(0))
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.utils.TermSelfReplace

    :param atom: a predicate or a term
    :param index: an index of a subterm to replace, must be greater than 0
    :param term: replacement term for a given index
    :returns:
    :raises NoSubtermFound: if subterm with a given index doesn't exist
    """
    subterm_length = 1
    if not isinstance(atom, Variable):
        for i, argument in enumerate(atom.arguments):
            if index == subterm_length:
                return dataclasses.replace(
                    atom,
                    arguments=atom.arguments[:i]
                    + (_replace_if_not_the_same(argument, term),)
                    + atom.arguments[i + 1 :],
                )
            try:
                return dataclasses.replace(
                    atom,
                    arguments=atom.arguments[:i]
                    + (
                        replace_subterm_by_index(
                            argument, index - subterm_length, term
                        ),
                    )
                    + atom.arguments[i + 1 :],
                )
            except NoSubtermFound as error:
                subterm_length += error.args[0]
    raise NoSubtermFound(subterm_length)


def _flat_list(list_of_lists: Tuple[Tuple[Any, ...], ...]) -> Tuple[Any, ...]:
    return tuple(set(chain(*list_of_lists)))


def reduce_to_proof(clauses: Dict[str, Clause]) -> Tuple[Clause, ...]:
    """
    Leave only clauses belonging to the refutational proof.

    >>> reduce_to_proof({
    ...     "one": Clause((), label="one"), "two": Clause((), label="two")
    ... })
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.utils.WrongRefutationProofError
    >>> state = {"one": Clause((), label="one")}
    >>> reduce_to_proof(state) == (Clause((), label="one"), )
    True

    :param clauses: a map of clause labels to clauses
    :returns: the reduced list of clauses
    :raises WrongRefutationProofError: if there is no complete refutation proof
        in a given proof state
    """
    empty_clauses = tuple(
        clause for clause in clauses.values() if clause.literals == tuple()
    )
    if len(empty_clauses) == 1:
        reduced: Tuple[Clause, ...] = ()
        new_reduced: Tuple[Clause, ...] = (empty_clauses[0],)
        while len(new_reduced) > 0:
            reduced += tuple(
                clause for clause in new_reduced if clause not in reduced
            )
            new_reduced = tuple(
                clauses[label]
                for label in tuple(
                    reversed(
                        sorted(
                            _flat_list(
                                tuple(
                                    (
                                        ()
                                        if clause.inference_parents is None
                                        else clause.inference_parents
                                    )
                                    for clause in new_reduced
                                )
                            )
                        )
                    )
                )
            )
        return reduced
    raise WrongRefutationProofError



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Agent Testing
==============

This module is an example of testing your own trained agent.
"""
import random
import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from operator import itemgetter
from typing import Any, Dict, List, Optional, Tuple

import gym
import orjson
from gym.wrappers import TimeLimit

from gym_saturation.envs.saturation_env import (
    STATE_DIFF_UPDATED,
    SaturationEnv,
)
from gym_saturation.logic_ops.utils import (
    WrongRefutationProofError,
    clause_length,
)


class BaseAgent:
    """A basic RL agent."""

    _state = None

    @property
    def state(self):
        """Agent can have its inner state."""
        return self._state

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        """
        Get an action given observations and other inputs.

        :param observation: an observation returned by the environment after
            the latest action (or it's initial state)
        :param reward: reward from the previous step
            (use zero for the first step)
        :param info: info dictionary returned by the environment
        :returns: an action
        """


@dataclass
class Transition:
    """
    An object describing environment's and agent's states.

    Before the agent's action, the action itself and its results
    """

    env_state: dict
    agent_state: Any
    action: int
    observation: dict
    reward: float
    done: bool
    info: Dict[str, Any]


class SizeAgent(BaseAgent):
    """
    Agent which selects the shortest clause.

    .. _size_agent:
    """

    def __init__(self):
        """Don't compute the formulae lengths twice."""
        self._state: Dict[str, Tuple[float, bool]] = {}

    def update_state(self, info: Dict[str, Any]) -> None:
        """
        Update the state of the agent according with the transition.

        :param info: an info dict (parm of environment response)
        """
        parsed_state_diff = tuple(map(orjson.loads, info[STATE_DIFF_UPDATED]))
        self._state.update(
            {
                clause["label"]: (clause_length(clause), clause["processed"])
                for clause in parsed_state_diff
            }
        )

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:  # noqa: D102
        self.update_state(info)
        return min(
            (
                (key, value[0])
                for key, value in enumerate(self.state.values())
                if not value[1]
            ),
            key=itemgetter(1),
        )[0]


class AgeAgent(BaseAgent):
    """
    Agent which selects the oldest clause.

    .. _age_agent:
    """

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:  # noqa: D102
        return min(
            i
            for i, clause in enumerate(
                map(orjson.loads, observation["real_obs"])
            )
            if not clause["processed"]
        )


class SizeAgeAgent(BaseAgent):
    """
    Agent taking several times the smallest clause and then the oldest.

    .. _size_age_agent:
    """

    def __init__(self, size_steps: int, age_steps: int):
        """
        Initialize two sub-agents.

        :param size_steps: how many times to select the shortest clause
        :param age_steps: how many times to select the oldest clause
        """
        self.size_steps = size_steps
        self.age_steps = age_steps
        self._step_count = 0
        self._use_size = True
        self._size_agent = SizeAgent()
        self._age_agent = AgeAgent()

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:  # noqa: D102
        self._step_count += 1
        if self._use_size:
            if self._step_count >= self.size_steps:
                self._step_count = 0
                self._use_size = False
            return self._size_agent.get_action(observation, reward, info)
        if self._step_count >= self.age_steps:
            self._step_count = 0
            self._use_size = True
        self._size_agent.update_state(info)
        return self._age_agent.get_action(observation, reward, info)


class RandomAgent(BaseAgent):
    """Agent which selects clauses randomly."""

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:  # noqa: D102
        return random.choice(
            [
                i
                for i, clause in enumerate(
                    map(orjson.loads, observation["real_obs"])
                )
                if not clause["processed"]
            ]
        )


def _proof_found_before_the_start(env: SaturationEnv) -> Tuple[float, bool]:
    try:
        _ = env.tstp_proof
        return 1.0, True
    except WrongRefutationProofError:
        return 0.0, False


def _reset_with_options(env: gym.Env) -> dict:
    res = env.reset()
    if isinstance(res, tuple):
        obs, _ = res  # pragma: nocover
    else:
        obs = res
    return obs


def episode(env: SaturationEnv, agent: BaseAgent) -> float:
    """
    Try to solve the problem and logs the clauses.

    >>> import os
    >>> import shutil
    >>> test_agent_output = "test_agent_output"
    >>> shutil.rmtree(test_agent_output, ignore_errors=True)
    >>> os.mkdir(test_agent_output)
    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> from glob import glob
    >>> problem_list = sorted(glob(os.path.join(
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST"
    ...     ))
    ... , "*-*.p")))
    >>> random.seed(0)
    >>> agents = (SizeAgeAgent(2, 1), SizeAgeAgent(1, 2), RandomAgent())
    >>> max_clauses = (100, 100, 4)
    >>> for i in range(3):
    ...     env = gym.make(
    ...         "GymSaturation-v0",
    ...         problem_list=[problem_list[i]],
    ...         max_clauses=max_clauses[i],
    ...     )
    ...     env._max_episode_steps = 5
    ...     agent_testing_report(env, agents[i])
    Proof of length 1 found in 2 steps:
    cnf(..., lemma, $false, inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).
    Step limit reached
    Proof state size limit reached

    :param env: a `gym_saturation` environment
    :param agent: an initialized agent. Must have `get_action` method
    :returns: total reward
    """
    obs = _reset_with_options(env)
    info: Dict[str, Any] = {STATE_DIFF_UPDATED: obs["real_obs"]}
    reward, done = _proof_found_before_the_start(env)
    total_reward = reward
    while not done:
        action = agent.get_action(obs, reward, info)
        obs, reward, done, info = env.step(action)
        total_reward += reward
    return total_reward


def parse_args(args: Optional[List[str]] = None) -> Namespace:
    """
    Parse script arguments.

    >>> parse_args([
    ...     "--problem_filename", "test",
    ...     "--step_limit", "1"
    ... ])
    Namespace(problem_filename='test', step_limit=1, vampire_binary_path=None)

    :param args: a list of string arguments
        (for testing and use in a non script scenario)
    :returns: arguments namespace for the script
    """
    argument_parser = ArgumentParser()
    argument_parser.add_argument("--problem_filename", type=str, required=True)
    argument_parser.add_argument("--step_limit", type=int, required=True)
    argument_parser.add_argument(
        "--vampire_binary_path", type=str, required=False
    )
    parsed_args = argument_parser.parse_args(args)
    return parsed_args


def agent_testing_report(env: SaturationEnv, agent: BaseAgent) -> None:
    """
    Print a report after testing an agent in an environment.

    :param env: an environment
    :param agent: an agent
    """
    total_reward = episode(env, agent)
    step_count = getattr(env, "_elapsed_steps")
    try:
        a_proof = env.tstp_proof
        if total_reward == 1.0:
            proof_length = len(a_proof.split("\n"))
            print(
                f"Proof of length {proof_length} found "
                f"in {step_count} steps:\n{a_proof}"
            )
    except WrongRefutationProofError:
        print(
            "Step limit reached"
            if step_count == getattr(env, "_max_episode_steps")
            else "Proof state size limit reached"
        )


def test_agent(args: Optional[List[str]] = None) -> None:
    """
    The main function for this module.

    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> import os
    >>> from glob import glob
    >>> problem_filenames = sorted(glob(os.path.join(
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST", "TST00*-1.p"
    ...     ))
    ... )))
    >>> test_agent([
    ...     "--problem_filename", problem_filenames[0],
    ...     "--step_limit", "3"
    ... ])
    Problem file: ...TST001-1.p
    Proof of length 1 found in 2 steps:
    cnf(..., lemma, $false, inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).
    >>> for problem_filename in problem_filenames:
    ...     test_agent([
    ...         "--problem_filename", problem_filename,
    ...         "--step_limit", "10",
    ...         "--vampire_binary_path", "vampire",
    ...     ])
    Problem file: ...TST001-1.p
    Proof of length 6 found in 0 steps:
    ...
    Problem file: ...TST002-1.p
    Proof of length 10 found in 4 steps:
    ...
    Problem file: ...TST003-1.p
    Proof of length 5 found in 3 steps:
    ...
    cnf(5, hyp..., $false, inference(subsumption_resolution, [], [4, 3])).

    :param args: parameters from the command line or explicitly set
    """
    sys.setrecursionlimit(10000)
    arguments = parse_args(args)
    env = (
        gym.make(
            "GymVampire-v0",
            problem_list=[arguments.problem_filename],
            vampire_binary_path=arguments.vampire_binary_path,
        )
        if arguments.vampire_binary_path is not None
        else gym.make(
            "GymSaturation-v0",
            problem_list=[arguments.problem_filename],
        )
    )
    environment = TimeLimit(env, arguments.step_limit)
    print(f"Problem file: {arguments.problem_filename}")
    agent_testing_report(environment, SizeAgeAgent(5, 1))  # type: ignore


if __name__ == "__main__":
    test_agent()  # pragma: no cover



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Vampire Wrapper
================
"""
from typing import Tuple

import pexpect


class VampireWrapper:
    """
    A wrapper around Vampire binary running in a manual clause selection mode.

    .. _vampire-wrapper :

    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> import os
    >>> tptp_folder = files("gym_saturation").joinpath(
    ...     os.path.join("resources", "TPTP-mock")
    ... )
    >>> tptp_problem = os.path.join(
    ...     tptp_folder, "Problems", "TST", "TST003-1.p"
    ... )
    >>> vampire = VampireWrapper("vampire")
    >>> vampire.pick_a_clause("2")
    Traceback (most recent call last):
     ...
    ValueError: start solving a problem first!
    >>> old_result = vampire.start(tptp_problem, tptp_folder)
    >>> print(old_result)
    (('input', '1', 'animal(X0) | ~carnivore(X0) [input]'),...
    >>> new_result = vampire.start(tptp_problem, tptp_folder)
    >>> old_result == new_result
    True
    >>> vampire.pick_a_clause("wrong")
    Traceback (most recent call last):
     ...
    ValueError:
    wrong
    ...
    """

    def __init__(self, binary_path: str):
        """
        Remember the path to Vampire, don't start the process yet.

        :param binary_path: an absolute path to Vampire binary
        """
        self.binary_path = binary_path
        self._proc = None

    def _get_stdout(self) -> Tuple[Tuple[str, str, str], ...]:
        result: Tuple[Tuple[str, str, str], ...] = ()
        self.proc.expect(["Pick a clause:", pexpect.EOF])
        for line in self.proc.before.decode("utf-8").split("\r\n"):
            if (
                line[:5] == "[SA] "
                or line[:12] in {"[PP] final: ", "[PP] input: "}
                or "[PP] fn def discovered: " in line
            ):
                result_type, result_body = line[5:].split(": ")
                clause_label, clause = result_body.split(". ")
                result += ((result_type, clause_label, clause),)
        if result:
            return result
        raise ValueError(self.proc.before.decode("utf-8"))

    def start(
        self, problem_filename: str, tptp_folder: str
    ) -> Tuple[Tuple[str, str, str], ...]:
        """
        Start Vampire in a manual mode on a given problem.

        Time limit is one day, Vampire prints everything

        :param problem_filename: full path of a TPTP problem file
        :param tptp_folder: the root folder for TPTP library
        :returns: a sequence of action type, clause number and clause
        """
        if self._proc is not None:
            self._proc.close()
        self._proc = pexpect.spawn(
            f"{self.binary_path} --manual_cs on --show_everything on "
            + "--time_limit 1D --avatar off "
            + f"--include {tptp_folder} {problem_filename}"
        )
        return self._get_stdout()

    def pick_a_clause(
        self, clause_label: str
    ) -> Tuple[Tuple[str, str, str], ...]:
        """
        Select a clause and get response from Vampire.

        :param clause_label: a given clause order number
        :returns: a sequence of action type, clause number and clause
        """
        self.proc.sendline(clause_label)
        return self._get_stdout()

    @property
    def proc(self) -> pexpect.spawn:
        """
        Vampire process.

        :raises ValueError: when called before ``reset``
        """
        if self._proc is None:
            raise ValueError("start solving a problem first!")
        return self._proc



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Unification
============
"""
from typing import Optional, Tuple

from tptp_lark_parser.grammar import Function, Predicate, Proposition, Variable

from gym_saturation.logic_ops.substitution import Substitution
from gym_saturation.logic_ops.utils import is_subproposition


class NonUnifiableError(Exception):
    """Exception raised when terms are not unifiable."""


def _get_disagreement(
    one: Proposition, two: Proposition
) -> Tuple[Proposition, ...]:
    """
    Find a disagreement set of two first order propositions.

    :param one: some proposition
    :param two: some (other but might be the same) proposition
    :returns: either an empty list (if there is no disagreement) or a pair of
        first subterms which disagree
    """
    if isinstance(one, Variable) and isinstance(two, Variable):
        if one.index != two.index:
            return (one, two)
        return ()
    if (
        isinstance(one, Predicate)
        and isinstance(two, Predicate)
        or isinstance(one, Function)
        and isinstance(two, Function)
    ):
        if one.index != two.index or len(one.arguments) != len(two.arguments):
            return (one, two)
        for argument_one, argument_two in zip(one.arguments, two.arguments):
            disagreement = _get_disagreement(argument_one, argument_two)
            if disagreement != ():
                return disagreement
        return ()
    return (one, two)


def _propose_substitution(
    disagreement: Tuple[Proposition, ...],
    propositions: Tuple[Proposition, ...],
) -> Substitution:
    substitution = None
    if isinstance(disagreement[0], Variable) and isinstance(
        disagreement[1], (Variable, Function)
    ):
        substitution = Substitution(disagreement[0], disagreement[1])
    elif isinstance(disagreement[1], Variable) and isinstance(
        disagreement[0], (Variable, Function)
    ):
        substitution = Substitution(disagreement[1], disagreement[0])
    if substitution is None or is_subproposition(
        substitution.variable, substitution.term
    ):
        raise NonUnifiableError((propositions[0], propositions[1]))
    return substitution


def most_general_unifier(
    propositions: Tuple[Proposition, ...],
    substitutions: Optional[Tuple[Substitution, ...]] = None,
) -> Tuple[Substitution, ...]:
    """
    Follow the Robinson's 1965 unification algorithm.

    >>> most_general_unifier((Variable(0), Variable(0)))
    ()
    >>> most_general_unifier((Variable(0), Function(7, (Variable(0),))))
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.unification.NonUnifiableError: (Variable(index=...
    >>> "this_is_a_test_case", most_general_unifier((Predicate(8, (Variable(0), Variable(1), Function(0, (Variable(2),)))), Predicate(8, (Variable(4), Function(7, (Variable(5), Variable(5))), Variable(5)))))
    ('this_is_a_test_case', (Substitution(variable=Variable(index=0), term=Variable(index=4)), Substitution(variable=Variable(index=1), term=Function(index=7, arguments=(Variable(index=5), Variable(index=5)))), Substitution(variable=Variable(index=5), term=Function(index=0, arguments=(Variable(index=2),)))))

    :param propositions: a set of propositions to unify
    :param substitutions: this function is recursive and uses this param to
        accumulate its output
    :returns: a map from variables to functions or other variables (aka
        substitution)
    """
    not_none_substitutions = () if substitutions is None else substitutions
    if len(propositions) == 1:
        return not_none_substitutions
    disagreement = _get_disagreement(propositions[0], propositions[1])
    if disagreement == ():
        return ()
    substitution = _propose_substitution(disagreement, propositions)
    return most_general_unifier(
        tuple(
            set(
                substitution(proposition)  # type: ignore
                for proposition in propositions
            )
        ),
        not_none_substitutions + (substitution,),
    )



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Saturation Environment with Vampire backend
============================================
"""
import dataclasses
import os
import random
from typing import Dict, List, Tuple, Union

import orjson
from tptp_lark_parser.grammar import Clause

from gym_saturation.envs.saturation_env import MAX_CLAUSES, SaturationEnv
from gym_saturation.vampire_wrapper import VampireWrapper


class VampireEnv(SaturationEnv):
    """
    An RL environment around a Vampire prover.

    This class has the same API as SaturationEnv but uses another backend.
    Here we have only a simple smoke test.
    For rigorous testing, see ``gym_saturation.agent_testing`` module.

    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> vampire_binary_path = files("gym_saturation").joinpath(
    ...     os.path.join("resources", "vampire-mock")
    ... )
    >>> vampire_env = VampireEnv(vampire_binary_path, ["test"])
    >>> vampire_env.reset()
    Traceback (most recent call last):
     ...
    ValueError: ('Unexpected reposnse type: ', 'who could expect that?')
    """

    def __init__(
        self,
        vampire_binary_path: str,
        problem_list: List[str],
        max_clauses: int = MAX_CLAUSES,
    ):
        """
        Initialize a :ref:`VampireWrapper <vampire-wrapper>`.

        :param vampire_binary_path: a path to Vampire binary
        :param problem_list: a list of names of TPTP problem files
        :param max_clauses: maximal number of clauses in proof state
        """
        super().__init__(problem_list, max_clauses)
        self._vampire = VampireWrapper(vampire_binary_path)

    def _parse_vampire_clause(
        self, clause_label: str, clause_text: str
    ) -> Clause:
        formula, inference_info = clause_text.split("[")
        pre_inference = inference_info.split("]")[0].split(" ")
        if len(pre_inference) > 1:
            inference_parents = ",".join(pre_inference[-1].split(","))
            inference_rule = "_".join(pre_inference[:-1])
        else:
            inference_parents, inference_rule = "", pre_inference[0]
        parsed_clause = self._tptp_parser.parse(
            f"cnf({clause_label}, hypothesis, ({formula}), "
            + f"inference({inference_rule}, [], [{inference_parents}]))."
        )[0]
        return dataclasses.replace(parsed_clause, processed=True)

    def _parse_vampire_reponse(
        self, vampire_response: Tuple[Tuple[str, str, str], ...]
    ) -> Dict[str, Clause]:
        updated: Dict[str, Clause] = {}
        for response_type, clause_label, clause_text in vampire_response:
            if response_type in {"new", "final", "input", "fn def discovered"}:
                updated[clause_label] = self._parse_vampire_clause(
                    clause_label, clause_text
                )
            elif response_type in {
                "active",
                "forward reduce",
                "passive",
                "backward reduce",
                "new propositional",
            }:
                changed_clause = dataclasses.replace(
                    self._state[clause_label]
                    if clause_label in self._state
                    else updated[clause_label],
                    processed=response_type
                    not in ("passive", "new propositional"),
                )
                updated[clause_label] = changed_clause
            else:
                raise ValueError("Unexpected reposnse type: ", response_type)
        return updated

    def reset(self) -> Union[dict, Tuple[dict, dict]]:  # noqa: D102
        self.problem = random.choice(self.problem_list)
        tptp_folder = os.path.join(os.path.dirname(self.problem), "..", "..")
        vampire_response = self._vampire.start(self.problem, tptp_folder)
        self._state = {}
        updated = self._parse_vampire_reponse(vampire_response)
        self._state = {
            clause.label: dataclasses.replace(clause, birth_step=0)
            for clause in updated.values()
        }
        return self.state

    def _do_deductions(self, action: int) -> Tuple[bytes, ...]:
        given_clause = list(self._state.values())[action]
        updated = self._parse_vampire_reponse(
            self._vampire.pick_a_clause(given_clause.label)
        )
        self._state.update(updated)
        return tuple(map(orjson.dumps, updated.values()))



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Factoring
==========
"""
from typing import Tuple

from tptp_lark_parser import grammar as gram

from gym_saturation.logic_ops.unification import (
    NonUnifiableError,
    most_general_unifier,
)


def factoring(
    given_clause: gram.Clause,
    literal_one: gram.Literal,
    literal_two: gram.Literal,
) -> gram.Clause:
    r"""
    Apply positive factoring rule.

    .. _factoring:

    .. math:: {\frac{C\vee A_1\vee A_2}{\sigma\left(C\vee L_1\right)}}

    where

    * :math:`C` and is a clause
    * :math:`A_1` and :math:`A_2` are atomic formulae (positive literals)
    * :math:`\sigma` is a most general unifier of :math:`A_1` and :math:`A_2`

    >>> from tptp_lark_parser.grammar import Predicate, Variable, Function
    >>> "this_is_a_test_case", factoring(gram.Clause((gram.Literal(True, Predicate(8, (Variable(0),))),)), gram.Literal(False, Predicate(9, (Variable(0),))), gram.Literal(False, Predicate(9, (Function(0, ()),)))).literals
    ('this_is_a_test_case', (Literal(negated=True, atom=Predicate(index=8, arguments=(Function(index=0, arguments=()),))), Literal(negated=False, atom=Predicate(index=9, arguments=(Function(index=0, arguments=()),)))))
    >>> factoring(
    ...     gram.Clause(()), gram.Literal(False, Predicate(8, ())),
    ...     gram.Literal(True, Predicate(7, ()))
    ... )
    Traceback (most recent call last):
     ...
    ValueError: factoring is not possible for Literal(negated=False, atom=P...

    :param given_clause: :math:`C`
    :param literal_one: :math:`A_1`
    :param literal_two: :math:`A_2`
    :returns: a new clause --- the factoring result
    :raises ValueError: if ``literal_one`` or ``literal_two`` is negated
    """
    if literal_one.negated or literal_two.negated:
        raise ValueError(
            f"factoring is not possible for {literal_one} and {literal_two}"
        )
    substitutions = most_general_unifier((literal_one.atom, literal_two.atom))
    new_literals = given_clause.literals + (literal_one,)
    result = gram.Clause(new_literals)
    for substitution in substitutions:
        result = substitution.substitute_in_clause(result)
    return result


def all_possible_factors(
    given_clause: gram.Clause,
) -> Tuple[gram.Clause, ...]:
    """
    One of the four basic building blocks of the Given Clause algorithm.

    >>> from tptp_lark_parser.tptp_parser import TPTPParser
    >>> parser = TPTPParser()
    >>> clause = parser.parse("cnf(one, axiom, p(c) | p(X) | q).")[0]
    >>> tuple(map(
    ...     parser.cnf_parser.pretty_print,
    ...     all_possible_factors(clause)
    ... ))
    ('cnf(x..., lemma, q() | p(c), inference(factoring, [], [one])).',)

    :param given_clause: a new clause which should be combined with all the
        processed ones
    :returns: results of all possible factors with each one from
        ``clauses`` and the ``given_clause``
    """
    factors: Tuple[gram.Clause, ...] = ()
    for i, literal_one in enumerate(given_clause.literals):
        for j in range(i + 1, len(given_clause.literals)):
            if (
                not literal_one.negated
                and not given_clause.literals[j].negated
            ):
                a_clause = gram.Clause(
                    given_clause.literals[:i]
                    + given_clause.literals[i + 1 : j]
                    + given_clause.literals[j + 1 :]
                )
                try:
                    factors = factors + (
                        factoring(
                            a_clause, literal_one, given_clause.literals[j]
                        ),
                    )
                except NonUnifiableError:
                    pass
    return tuple(
        gram.Clause(
            literals=factor.literals,
            inference_parents=(given_clause.label,),
            inference_rule="factoring",
        )
        for ord_num, factor in enumerate(factors)
    )



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Substitution
=============
"""
import dataclasses
from typing import Union

from tptp_lark_parser import grammar


@dataclasses.dataclass
class Substitution:
    """
    A mapping from ``Variable`` to ``Term``.

    >>> substitution = Substitution(
    ...     grammar.Variable(0), grammar.Function(2, ())
    ... )
    >>> substitution(
    ...     grammar.Clause((
    ...         grammar.Literal(
    ...             False,
    ...             grammar.Predicate(
    ...                 3,
    ...                 (grammar.Function(1, (grammar.Variable(0),)),))
    ...             ),
    ...     ), label="this_is_a_test_case")
    ... )
    Clause(literals=(Literal(negated=False, atom=Predicate(index=3, arguments=(Function(index=1, arguments=(Function(index=2, arguments=()),)),))),), label='this_is_a_test_case', role='lemma', inference_parents=None, inference_rule=None, processed=None, birth_step=None)
    >>> substitution(grammar.Variable(0))
    Function(index=2, arguments=())
    """

    variable: grammar.Variable
    term: grammar.Term

    def __call__(
        self, target: Union[grammar.Clause, grammar.Proposition]
    ) -> Union[grammar.Clause, grammar.Proposition]:
        """
        Apply a substitution to a given target clause.

        :param target: a clause to which to apply the substitution
        """
        if isinstance(target, grammar.Clause):
            return self.substitute_in_clause(target)
        if isinstance(target, grammar.Predicate):
            return self._substitute_in_predicate(target)
        return self._substitute_in_term(target)

    def _substitute_in_term(self, term: grammar.Term) -> grammar.Term:
        if isinstance(term, grammar.Function):
            return grammar.Function(
                term.index,
                tuple(
                    self._substitute_in_term(argument)
                    for argument in term.arguments
                ),
            )
        if term.index == self.variable.index:
            return self.term
        return term

    def _substitute_in_predicate(
        self, predicate: grammar.Predicate
    ) -> grammar.Predicate:
        return grammar.Predicate(
            predicate.index,
            tuple(
                self._substitute_in_term(argument)
                for argument in predicate.arguments
            ),
        )

    def substitute_in_clause(self, clause: grammar.Clause) -> grammar.Clause:
        """
        Apply the substitution to something which is known to be a ``Clause``.

        :param clause: a clause to apply substitution to
        :returns: the result of the substitution
        """
        literals = tuple(
            grammar.Literal(
                literal.negated,
                self._substitute_in_predicate(literal.atom),
            )
            for literal in clause.literals
        )
        return dataclasses.replace(clause, literals=literals)



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Clause Space
=============
"""
import orjson
from gym import spaces


class ClauseSpace(spaces.Space):
    """
    An OpenAI Gym space for a list of bytes.

    .. _clause_space:

    >>> space = ClauseSpace()
    >>> space.sample() in space
    True
    >>> "no" in space
    False
    >>> ["no"] in space
    False
    """

    def __init__(self):  # noqa: D107
        super().__init__(shape=(0,))

    def contains(self, x):  # noqa: D102
        if isinstance(x, list):
            for byte_sequence in x:
                return isinstance(byte_sequence, bytes)
        return False

    def sample(self):  # noqa: D102
        return [
            orjson.dumps(
                {
                    "literals": (),
                    "label": "test",
                    "role": "lemma",
                    "birth_step": 0,
                    "processed": False,
                    "inference_parents": (),
                    "inference_rule": "",
                }
            )
        ]



# Copyright 2021-2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa: D205, D400
"""
Reflexivity Resolution
=======================
"""
from typing import Tuple

from tptp_lark_parser import grammar

from gym_saturation.logic_ops.unification import (
    NonUnifiableError,
    most_general_unifier,
)


def reflexivity_resolution(
    given_clause: grammar.Clause, a_literal: Tuple[grammar.Term, grammar.Term]
) -> grammar.Clause:
    r"""
    Apply reflexivity resolution rule.

    .. _reflexivity_resolution:

    .. math:: \frac{C\vee s\not\approx t}{\sigma\left(C\right)}

    where

    * :math:`C` and is a clause
    * :math:`s` and :math:`t` are terms, :math:`\not\approx` is a negation of equality
    * :math:`\sigma` is a most general unifier of :math:`s` and :math:`t`

    >>> from tptp_lark_parser.grammar import Predicate, Variable, Function
    >>> "this_is_a_test_case", reflexivity_resolution(grammar.Clause((grammar.Literal(True, Predicate(7, (Variable(0),))),)), (Variable(0), Function(0, ()))).literals
    ('this_is_a_test_case', (Literal(negated=True, atom=Predicate(index=7, arguments=(Function(index=0, arguments=()),))),))

    :param given_clause: :math:`C`
    :param a_literal: :math:`s\not\approx t`
    :returns: a new clause --- the reflexivity resolution result
    """
    substitutions = most_general_unifier((a_literal[0], a_literal[1]))
    new_literals = given_clause.literals
    result = grammar.Clause(new_literals)
    for substitution in substitutions:
        result = substitution.substitute_in_clause(result)
    return result


def all_possible_reflexivity_resolvents(
    given_clause: grammar.Clause,
) -> Tuple[grammar.Clause, ...]:
    """
    One of the four basic building blocks of the Given Clause algorithm.

    >>> from tptp_lark_parser.tptp_parser import TPTPParser
    >>> parser = TPTPParser()
    >>> clause = parser.parse("cnf(this_is_a_test_case, axiom, p(X) | ~ X=a | b != a).")[0]
    >>> tuple(map(
    ...     parser.cnf_parser.pretty_print,
    ...     all_possible_reflexivity_resolvents(clause)
    ... ))
    ('cnf(x..., lemma, p(a) | ~b = a, inference(reflexivity_resolution, [], [this_is_a_test_case])).',)

    :param given_clause: a new clause which should be combined with all the
        processed ones
    :returns: results of all possible reflexivity resolvents with each one from
        ``clauses`` and the ``given_clause``
    """
    reflexivity_resolvents: Tuple[grammar.Clause, ...] = ()
    for i, a_literal in enumerate(given_clause.literals):
        if (
            a_literal.negated
            and a_literal.atom.index == grammar.EQUALITY_SYMBOL_ID
            or not a_literal.negated
            and a_literal.atom.index == grammar.INEQUALITY_SYMBOL_ID
        ) and len(a_literal.atom.arguments) == 2:
            a_clause = grammar.Clause(
                given_clause.literals[:i] + given_clause.literals[i + 1 :]
            )
            if a_clause.literals:
                try:
                    reflexivity_resolvents = reflexivity_resolvents + (
                        reflexivity_resolution(
                            a_clause,
                            (
                                a_literal.atom.arguments[0],
                                a_literal.atom.arguments[1],
                            ),
                        ),
                    )
                except NonUnifiableError:
                    pass
    return tuple(
        grammar.Clause(
            literals=reflexivity_resolvent.literals,
            inference_parents=(given_clause.label,),
            inference_rule="reflexivity_resolution",
        )
        for ord_num, reflexivity_resolvent in enumerate(reflexivity_resolvents)
    )
